/******************************************************************************
 * Copyright 2004, 2005 Alex Karev.
 *
 * This file is part of gtkpivo library.
 *
 * gtkpivo library is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation version 2.1 of the License.
 *
 * gtkpivo library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with gtkpivo library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 ******************************************************************************/
/* vim: set ts=4 sw=4 : */
#include "gtkled.h"

#include <string.h>

#define LABEL_FONT "Sans Bold 9"
#define VAL_FONT "Sans 7"

#define VAL_LABEL "1 KVA QUA 1310"

#define LED_DEFAULT_WIDTH	20
#define LED_DEFAULT_HEIGHT	20

static GtkWidgetClass *parent_class = NULL;

static void gtk_led_class_init(GtkLedClass *class);
static void gtk_led_init(GtkLed *led);
static void gtk_led_destroy(GtkObject *object);
static void gtk_led_realize(GtkWidget *widget);
static void gtk_led_size_request(GtkWidget      *widget,
				      GtkRequisition *requisition);
static void gtk_led_size_allocate(GtkWidget     *widget,
					   GtkAllocation *allocation);
static gboolean gtk_led_expose(GtkWidget      *widget,
				GdkEventExpose *event );

static gboolean draw_led_dynamic(GtkLed *led);
static gboolean draw_led_static(GtkLed *led);

static void adj_val_changed(GtkAdjustment *adj, gpointer arg);
static void adj_changed(GtkAdjustment *adj, gpointer arg);
static  void remove_adj(GtkLed *led);

static void
gtk_led_class_init(GtkLedClass *class)
{
	GtkObjectClass *object_class;
	GtkWidgetClass *widget_class;
	
	object_class = (GtkObjectClass*) class;
	widget_class = (GtkWidgetClass*) class;

	parent_class = gtk_type_class (gtk_widget_get_type ());

	object_class->destroy			= gtk_led_destroy;
	
	widget_class->size_request		= gtk_led_size_request;
	widget_class->realize			= gtk_led_realize;
	widget_class->size_allocate		= gtk_led_size_allocate;
	widget_class->expose_event		= gtk_led_expose;
}

static void
gtk_led_init(GtkLed *led)
{
	led->font_label = pango_font_description_from_string(LABEL_FONT);
	led->font_val = pango_font_description_from_string(VAL_FONT);

	led->adj	= NULL;
	led->label	= NULL;
	led->state  = GTK_LED_OFF;
}


static void
gtk_led_destroy(GtkObject *object)
{
	GtkLed *led;

	g_return_if_fail (object != NULL);
	g_return_if_fail (GTK_IS_LED (object));

	led = GTK_LED (object);
	
	if (led->label != NULL)
		g_free(led->label);
	led->label = NULL;

	if (led->adj != NULL) 
		remove_adj(led);
	led->adj = NULL;

	if (GTK_OBJECT_CLASS (parent_class)->destroy)
		(* GTK_OBJECT_CLASS (parent_class)->destroy) (object);
}

static void
gtk_led_realize(GtkWidget *widget)
{
	GtkLed *led;
	GdkWindowAttr attributes;
	gint attributes_mask;

	g_return_if_fail (widget != NULL);
	g_return_if_fail (GTK_IS_LED(widget));

	GTK_WIDGET_SET_FLAGS (widget, GTK_REALIZED);
	led = GTK_LED (widget);

	attributes.x		= widget->allocation.x;
	attributes.y		= widget->allocation.y;
	attributes.width	= widget->allocation.width;
	attributes.height	= widget->allocation.height;

	attributes.wclass		= GDK_INPUT_OUTPUT;
	attributes.window_type	= GDK_WINDOW_CHILD;
	attributes.event_mask	= gtk_widget_get_events (widget) | 
		GDK_EXPOSURE_MASK /*| GDK_STRUCTURE_MASK */;

	attributes.visual		= gtk_widget_get_visual (widget);
	attributes.colormap		= gtk_widget_get_colormap (widget);

   	attributes_mask = GDK_WA_X | GDK_WA_Y | GDK_WA_VISUAL | GDK_WA_COLORMAP;
	
	widget->window = gdk_window_new( gtk_widget_get_parent_window(widget), 
			&attributes, 
			attributes_mask);
	gdk_window_set_user_data(widget->window, led);

#if 1
	gtk_widget_modify_bg(widget, GTK_STATE_NORMAL, 
		&widget->style->white );
	gtk_widget_modify_fg(widget, GTK_STATE_ACTIVE, 
		&widget->style->black );
#endif
	
	widget->style = gtk_style_attach(widget->style, widget->window);
	gtk_style_set_background(widget->style, widget->window, GTK_STATE_NORMAL);
}

static void
gtk_led_size_allocate(GtkWidget     *widget,
					   GtkAllocation *allocation)
{
	g_return_if_fail (widget != NULL);
	g_return_if_fail (GTK_IS_LED (widget));
	g_return_if_fail (allocation != NULL);

	widget->allocation = *allocation;
	(* GTK_WIDGET_CLASS (parent_class)->size_allocate) (widget, allocation);
	
	if (GTK_WIDGET_REALIZED(widget)) {
		gdk_window_move_resize(widget->window,
				allocation->x, allocation->y,
				allocation->width, allocation->height);

	}
}


static void 
gtk_led_size_request(GtkWidget      *widget,
				      GtkRequisition *requisition)
{
	requisition->width		= LED_DEFAULT_WIDTH;
	requisition->height		= LED_DEFAULT_HEIGHT;
}

static gboolean
gtk_led_expose(GtkWidget      *widget,
				GdkEventExpose *event )
{
	GtkLed *led = GTK_LED(widget);

	gint width, height;
	
    width  = widget->allocation.width;
    height = widget->allocation.height;

	if (event->count > 0)
		return (FALSE);

	(void) draw_led_dynamic(led);

	gtk_paint_shadow( widget->style, widget->window,
			widget->state, GTK_SHADOW_ETCHED_OUT,
			NULL, widget,  "led",
			0, 0,
			width, height );

	return (FALSE);
}


static gboolean
draw_led_dynamic(GtkLed *led)
{
	GdkGC *pgc;
	gint 
		x,y, width, height;

#if 1	
	gchar label[32];
	gint 
		dx, dy;

	g_snprintf(label, sizeof label, "%3.2f", led->adj->value );

	led->layout_val = gtk_widget_create_pango_layout(GTK_WIDGET(led), label);
	pango_layout_set_font_description (led->layout_val, led->font_val); 
	pango_layout_get_pixel_size(led->layout_val, &dx, &dy);

	dx = ( GTK_WIDGET(led)->allocation.width  - dx ) / 2;
	dy = ( GTK_WIDGET(led)->allocation.height - dy ) / 2;

	gdk_draw_layout( GTK_WIDGET(led)->window,
				GTK_WIDGET(led)->style->black_gc,
				dx , dy,
				led->layout_val );
	
	g_object_unref(led->layout_val);
#endif
	
	switch(led->state) {
		case GTK_LED_OFF:
			pgc = GTK_WIDGET(led)->style->black_gc;
			break;
		case GTK_LED_ON:
			pgc = GTK_WIDGET(led)->style->white_gc;
			break;
		default:
			pgc = NULL;
			break;
	}
	

	x = 0, y = 0; 
	
	width	= GTK_WIDGET(led)->allocation.width;
	height	= GTK_WIDGET(led)->allocation.height;


	
	if (NULL != pgc) {
		gdk_draw_arc( GTK_WIDGET(led)->window, 
				GTK_WIDGET(led)->style->black_gc, TRUE,
				x, y, width, height,
				0, 0 /* gtk_adjustment_get_value(led->adj) */);
	}

	return (TRUE);
}

static gboolean
draw_led_static(GtkLed *led)
{
	g_return_val_if_fail(led != NULL, FALSE);

	return (TRUE);
}

static void
adj_changed(GtkAdjustment *adj, gpointer arg)
{
	GtkLed *led = (GtkLed *) arg;

	g_return_if_fail(led != NULL);
}

static void
adj_val_changed(GtkAdjustment *adj, gpointer arg)
{

	GtkLed *led = (GtkLed *) arg;
	
	g_assert(led != NULL);

	led->state = (adj->value > 0) ? GTK_LED_ON : GTK_LED_OFF;
	gtk_widget_queue_draw(GTK_WIDGET(led));
}

static  void
remove_adj(GtkLed *led)
{
	g_return_if_fail(led != NULL);
	g_return_if_fail(led->adj != NULL);
	
	g_signal_handlers_disconnect_by_func( G_OBJECT(led->adj),
			G_CALLBACK(adj_changed), 
			led);
	g_signal_handlers_disconnect_by_func( G_OBJECT(led->adj),
			G_CALLBACK(adj_val_changed), 
			led);

	g_object_unref(led->adj);
}


/** 
 *	
 *	Public functions
 * 
 */
GtkType
gtk_led_get_type(void)
{
	static GtkType led_type = 0;

	if (!led_type)
	{
		static const GtkTypeInfo led_info = 
		{
			"GtkLed",
			sizeof (GtkLed),
			sizeof (GtkLedClass),
			(GtkClassInitFunc) gtk_led_class_init,
			(GtkObjectInitFunc) gtk_led_init,
			/* reserved_1 */ NULL,
			/* reserved_1 */ NULL,
			(GtkClassInitFunc) NULL
		};

		led_type = gtk_type_unique (GTK_TYPE_WIDGET, &led_info);
	}
	
	return (led_type);
}

GtkWidget* 
gtk_led_new(GtkAdjustment *adj, const gchar *label )
{
	GtkLed *led;
		
	led = gtk_type_new(gtk_led_get_type());

	/** Adjustment setup */
	gtk_led_set_adjustment(led, adj);

#if 0
	/** Static labels setup */
	setup_static_text(led);
#endif
	
	/** Creating label */
	if (label == NULL) {
		led->label = g_strndup(VAL_LABEL, strlen(VAL_LABEL));
	} else {
		led->label = g_strndup(label, 32);
	}

	return GTK_WIDGET(led);
}

void
gtk_led_set_adjustment(GtkLed *led, GtkAdjustment *adj)
{
	g_return_if_fail( led != NULL );
	g_return_if_fail( GTK_IS_LED(led) );

	if (adj) 
		g_return_if_fail(GTK_IS_ADJUSTMENT(adj));
	else 
		adj = (GtkAdjustment *) gtk_adjustment_new(0, 0., 999., 0.001, 1, 10);

	
	if (led->adj != adj) {
		if (led->adj != NULL)
			remove_adj(led);


		led->adj = adj;
		if (adj) {
			g_object_ref(adj);
			gtk_object_sink(GTK_OBJECT(adj));

			g_signal_connect(G_OBJECT(adj), "changed", G_CALLBACK(adj_changed),
				led);
			g_signal_connect(G_OBJECT(adj), "value-changed", G_CALLBACK(adj_val_changed),
				led);
		}
	}
}

