/******************************************************************************
 * Copyright 2004, 2005 Alex Karev.
 *
 * This file is part of gtkpivo library.
 *
 * gtkpivo library is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation version 2.1 of the License.
 *
 * gtkpivo library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with gtkpivo library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 ******************************************************************************/
/* vim: set ts=4 sw=4 : */

#define _USE_MATH_DEFINES
#define M_PI       3.14159265358979323846
#include <math.h>
#include <string.h>
#include "gtktaho.h"


#define TAHO_DEFAULT_WIDTH 	120  
#define TAHO_DEFAULT_HEIGHT 150  

#define TAHO_JIFFIES 46

#define RED_LB 5
#define RED_HB 5

#define YELLOW_LB 5
#define YELLOW_HB 5

#define RAD_START_OFFSET (M_PI / 2.)
#define RAD_OFFSET       (M_PI / 180. * 90)
#define RAD_END_OFFSET   ( RAD_START_OFFSET + RAD_OFFSET)

#define RIGHT_OFFSET	30
#define LEFT_OFFSET		10
#define TOP_OFFSET		10
#define BUTTOM_OFFSET	40

#define LONG_TICKS_LINE    20
#define SHORT_TICKS_LINE   12

#define LABEL_FONT "Sans 6"
#define VAL_FONT "Sans Bold 8"

#define VAL_NAME "Val"
#define VAL_SFX "mA"

static GtkWidgetClass *parent_class = NULL;

static void gtk_taho_class_init(GtkTahoClass *class);
static void gtk_taho_init(GtkTaho *taho);
static void gtk_taho_destroy (GtkObject *object);
static void gtk_taho_realize(GtkWidget *widget);
static void gtk_taho_size_request(GtkWidget      *widget,
				      GtkRequisition *requisition);
static void gtk_taho_size_allocate(GtkWidget     *widget,
					   GtkAllocation *allocation);
static gboolean gtk_taho_expose(GtkWidget      *widget,
				GdkEventExpose *event );

static void adj_changed(GtkAdjustment *adj, gpointer arg);
static void adj_val_changed(GtkAdjustment *adj, gpointer arg);
static  void remove_adj(GtkTaho *t);

static void draw_taho_dynamic(GtkTaho *t);
static void setup_static_text(GtkTaho *taho);

GtkType
gtk_taho_get_type(void)
{
	static GtkType taho_type = 0;

	if (!taho_type)
	{
		static const GtkTypeInfo taho_info = 
		{
			"GtkTaho",
			sizeof (GtkTaho),
			sizeof (GtkTahoClass),
			(GtkClassInitFunc) gtk_taho_class_init,
			(GtkObjectInitFunc) gtk_taho_init,
			/* reserved_1 */ NULL,
			/* reserved_1 */ NULL,
			(GtkClassInitFunc) NULL
		};

		taho_type = gtk_type_unique (GTK_TYPE_WIDGET, &taho_info);
	}
	
	return (taho_type);
}

static void
gtk_taho_class_init(GtkTahoClass *class)
{
	GtkObjectClass *object_class;
	GtkWidgetClass *widget_class;
	
	object_class = (GtkObjectClass*) class;
	widget_class = (GtkWidgetClass*) class;

	parent_class = gtk_type_class (gtk_widget_get_type ());

	object_class->destroy = gtk_taho_destroy;
	
	widget_class->realize = gtk_taho_realize;
	widget_class->expose_event = gtk_taho_expose;
	widget_class->size_request = gtk_taho_size_request;
	widget_class->size_allocate = gtk_taho_size_allocate;
}

static void
gtk_taho_init(GtkTaho *taho)
{
	taho->angle = RAD_END_OFFSET;
	taho->radius = 0.0;

	taho->xc = taho->yc = 0;

	taho->font_label = pango_font_description_from_string (LABEL_FONT);
	taho->font_val = pango_font_description_from_string (VAL_FONT);

	taho->label = taho->val_suffix = NULL;
}

void
gtk_taho_set_adjustment(GtkTaho *t, GtkAdjustment *adj)
{
	g_return_if_fail( t != NULL );
	g_return_if_fail( GTK_IS_TAHO(t) );
	
	if (adj) 
		g_return_if_fail(GTK_IS_ADJUSTMENT(adj));
	else 
		adj = (GtkAdjustment *) gtk_adjustment_new(0, 0., 999., 0.001, 1, 10);

	if (t->adj != adj) {
	
		if (t->adj != NULL)
			remove_adj(t);
	

		t->adj = adj;

		if (adj != NULL) {
			g_object_ref(adj);
			gtk_object_sink(GTK_OBJECT(adj));

			g_signal_connect(G_OBJECT(adj), "changed", G_CALLBACK(adj_changed),
				t);
			g_signal_connect(G_OBJECT(adj), "value-changed", G_CALLBACK(adj_val_changed),
				t);
		}
	}
}

GtkAdjustment*	
gtk_taho_get_adjustment(GtkTaho *taho)
{
	g_return_val_if_fail(taho != NULL, NULL);
	return (taho->adj);
}

GtkWidget*
gtk_taho_new(GtkAdjustment *adj, const gchar *label, const gchar *val_suffix)
{
	GtkTaho *taho;
		
	taho = gtk_type_new(gtk_taho_get_type());


	gtk_taho_set_adjustment(taho, adj);

	/** Static labels setup */
	setup_static_text(taho);
	
	if (label == NULL) {
		taho->label = g_strndup(VAL_NAME, strlen(VAL_NAME));
	} else {
		taho->label = g_strndup(label, 16);
	}

	if (val_suffix == NULL) {
		taho->val_suffix = g_strndup(VAL_SFX, strlen(VAL_SFX));
	} else {
		g_free(taho->val_suffix);
		taho->val_suffix = g_strndup(val_suffix, 16);
	}

	return GTK_WIDGET(taho);
}

static void
gtk_taho_destroy (GtkObject *object)
{
	GtkTaho *taho;

	g_return_if_fail (object != NULL);
	g_return_if_fail (GTK_IS_TAHO (object));

	taho = GTK_TAHO (object);
	
	if (taho->label != NULL)
		g_free(taho->label);

	taho->label = NULL;
	
	if (taho->val_suffix != NULL)
		g_free(taho->val_suffix);

	taho->val_suffix = NULL;


	if (taho->adj != NULL) 
		remove_adj(taho);

	taho->adj = NULL;
	
#if 0
	if (taho->red_gc)
		g_object_unref(GTK_OBJECT(taho->red_gc));

	if (taho->green_gc)
		g_object_unref(GTK_OBJECT(taho->green_gc));

	if (taho->yellow_gc)
		g_object_unref(GTK_OBJECT(taho->yellow_gc));

	if (taho->font_label)
		pango_font_description_free (taho->font_label);
	
	if (taho->font_val)
		pango_font_description_free (taho->font_val);

	if (taho->adj)
		g_object_unref(taho->adj);

#endif

	if (GTK_OBJECT_CLASS(parent_class)->destroy)
		(* GTK_OBJECT_CLASS (parent_class)->destroy) (object);
}

static void 
setup_static_text(GtkTaho *taho)
{
	gchar txt_upper[32], txt_lower[32];

	g_return_if_fail(taho != NULL);
	g_return_if_fail(taho->adj != NULL);

	if (taho->layout_label_upper != NULL)
		g_object_unref(taho->layout_label_upper);

	g_snprintf(txt_upper, sizeof txt_upper, "%2.2f", taho->adj->upper);
	taho->layout_label_upper = gtk_widget_create_pango_layout(GTK_WIDGET(taho), txt_upper);
	pango_layout_set_font_description (taho->layout_label_upper, taho->font_label); 
	

	if (taho->layout_label_lower != NULL)
		g_object_unref(taho->layout_label_lower);

	g_snprintf(txt_lower, sizeof txt_lower, "%2.2f", taho->adj->lower);
	taho->layout_label_lower = gtk_widget_create_pango_layout(GTK_WIDGET(taho), txt_lower);
	pango_layout_set_font_description (taho->layout_label_lower, taho->font_label); 
}

static void 
draw_taho_dynamic(GtkTaho *t)
{
	gfloat c, s;
	/* GdkPoint m[4]; */
	GdkPoint m[2]; 
	gchar txt[32];
	gint dx, dy;

	g_return_if_fail(t != NULL);
	g_return_if_fail(t->bgpix != NULL);

	/** FIXME: write fast computations */
	c = cos(t->angle);
	s = sin(t->angle);

	m[0].x = t->xc + c * (t->radius - 4);
	m[0].y = t->yc - s * (t->radius - 4);

	m[1].x = t->xc /* + cos(t->angle + M_PI) * 8 */;
	m[1].y = t->yc /* - sin(t->angle + M_PI) * 8 */;

	/*
	m[0].x = t->xc + c * (t->radius + 4);
	m[0].y = t->yc - s * (t->radius + 4);

	m[1].x = t->xc - 4 + cos(pheta - M_PI / 180. * 140.) * 8;
	m[1].y = t->yc - 4 - sin(pheta - M_PI / 180. * 140.) * 8;
	
	m[2].x = t->xc - 4;
	m[2].y = t->yc - 4;
	
	m[3].x = t->xc - 4 + cos(pheta + M_PI / 180. * 140.) * 8;
	m[3].y = t->yc - 4 - sin(pheta + M_PI / 180. * 140.) * 8;
	*/

	/** Drawing pointer */
	gdk_draw_line(GTK_WIDGET(t)->window,
				t->line_gc,
				m[0].x,	m[0].y, 
				m[1].x, m[1].y);

	/*
	gdk_draw_polygon( GTK_WIDGET(t)->window,
				GTK_WIDGET(t)->style->white_gc,
				TRUE,
				m, G_N_ELEMENTS(m));
	*/

	/** Draw val text */
	g_snprintf(txt, sizeof txt, "%s: %2.2f %s", t->label, t->adj->value, t->val_suffix);
	t->layout_val = gtk_widget_create_pango_layout(GTK_WIDGET(t), txt);
	pango_layout_set_font_description (t->layout_val, t->font_val); 
	pango_layout_get_pixel_size(t->layout_val, &dx, &dy);

	gdk_draw_layout (GTK_WIDGET(t)->window,
				GTK_WIDGET(t)->style->white_gc,
				(GTK_WIDGET(t)->allocation.width - dx) / 2, 
				t->yc + (BUTTOM_OFFSET + dy) / 2 - dy, 
				t->layout_val );
	
	g_object_unref(t->layout_val);
}

static void
draw_taho_bg_pix(GtkTaho *t)
{
	gint width, height;
	GdkGC *line_gc;
	/* GdkGC *gc; */
	/* GdkColor color; */
	int i, line_length = 0;
	gfloat pheta, c, s;

	g_return_if_fail (t != NULL);
	g_return_if_fail (t->bgpix != NULL);

	/*
	gc = gdk_gc_new(t->bgpix);
	gdk_color_parse("dark blue", &color);
	gdk_gc_set_rgb_fg_color(gc,&color);
	*/

	/* Fill background */
	gdk_drawable_get_size(GDK_DRAWABLE(t->bgpix), &width, &height);
	
	gdk_draw_rectangle(t->bgpix,
			GTK_WIDGET(t)->style->black_gc /* gc */,
			TRUE,
			0, 0,
			width, 
			height );

	for(i = 0; i < TAHO_JIFFIES; i++) {
		pheta = (i * M_PI * 0.0111 + RAD_START_OFFSET);
		c = cos(pheta);
		s = sin(pheta);
		
		line_length = (i % 5) ? SHORT_TICKS_LINE : LONG_TICKS_LINE ;
		
		if ( i <= RED_LB || i >= (TAHO_JIFFIES - RED_HB - 1) ) {
			line_gc = t->red_gc;
		} else if (i <= (RED_LB + YELLOW_LB) || i >= (TAHO_JIFFIES - (RED_HB + YELLOW_HB) -1) ) {
			line_gc = t->yellow_gc;
		} else {
			line_gc = t->green_gc;
		}
		
		gdk_draw_line( t->bgpix,
				/* GTK_WIDGET(t)->style->white_gc */line_gc,
				t->xc + c * t->radius,	
				t->yc - s * t->radius,
				t->xc + c * (t->radius - line_length), 
				t->yc - s * (t->radius - line_length) );
	}
	
	/** drawing labels */
	gdk_draw_layout (t->bgpix,
				GTK_WIDGET(t)->style->white_gc /* gc */,
				t->xc + cos(M_PI) * t->radius - 2, 
				t->yc - sin(M_PI) * t->radius + 2, 
				t->layout_label_lower);
	
	gdk_draw_layout (t->bgpix,
				GTK_WIDGET(t)->style->white_gc /* gc */,
				t->xc + cos(M_PI / 2.) * t->radius + 4, 
				t->yc - sin(M_PI / 2.) * t->radius, 
				t->layout_label_upper);

	/*
	g_object_unref(gc);
	*/
}

static void
gtk_taho_realize(GtkWidget *widget)
{
	GtkTaho *taho;
	GdkColor c;
	GdkWindowAttr attributes;
	gint attributes_mask;

	g_return_if_fail (widget != NULL);
	g_return_if_fail (GTK_IS_TAHO (widget));

	GTK_WIDGET_SET_FLAGS (widget, GTK_REALIZED);
	taho = GTK_TAHO (widget);

	attributes.x = widget->allocation.x;
	attributes.y = widget->allocation.y;
	attributes.width = widget->allocation.width;
	attributes.height = widget->allocation.height;

	attributes.wclass = GDK_INPUT_OUTPUT;
	attributes.window_type = GDK_WINDOW_CHILD;
	attributes.event_mask = gtk_widget_get_events (widget) | 
		GDK_EXPOSURE_MASK /*| GDK_STRUCTURE_MASK */;

	attributes.visual = gtk_widget_get_visual (widget);
	attributes.colormap = gtk_widget_get_colormap (widget);

   	attributes_mask = GDK_WA_X | GDK_WA_Y | GDK_WA_VISUAL | GDK_WA_COLORMAP;
	
	widget->window = gdk_window_new(gtk_widget_get_parent_window(widget), 
			&attributes, 
			attributes_mask);

	widget->style = gtk_style_attach(widget->style, widget->window);
	
	gdk_window_set_user_data(widget->window, widget);
	gtk_style_set_background(widget->style, widget->window, GTK_STATE_ACTIVE);
	
	/* Setup GCs */
	taho->red_gc = gdk_gc_new(widget->window);
	gdk_color_parse("red", &c);
	gdk_gc_set_rgb_fg_color(taho->red_gc, &c);

	taho->green_gc = gdk_gc_new(widget->window);
	gdk_color_parse("green", &c);
	gdk_gc_set_rgb_fg_color(taho->green_gc, &c);
	
	taho->yellow_gc = gdk_gc_new(widget->window);
	gdk_color_parse("yellow", &c);
	gdk_gc_set_rgb_fg_color(taho->yellow_gc, &c);

	taho->line_gc = gdk_gc_new(widget->window);
	gdk_color_parse("white", &c);
	gdk_gc_set_rgb_fg_color(taho->line_gc, &c);
	gdk_gc_set_line_attributes(taho->line_gc,
			3, 
			GDK_LINE_SOLID,
			GDK_CAP_ROUND,
			GDK_JOIN_ROUND);


	/** Setup bg pixmap */
	if (taho->bgpix != NULL) {
		gdk_window_set_back_pixmap(widget->window, NULL, FALSE);
		g_object_unref(taho->bgpix);
	}

	taho->bgpix = gdk_pixmap_new(widget->window, 
			widget->allocation.width, widget->allocation.height,
			-1);
	draw_taho_bg_pix(taho);
	gdk_window_set_back_pixmap(widget->window, taho->bgpix, FALSE);

	/*
	g_object_ref(taho->bgpix);
	*/

	draw_taho_dynamic(taho);
#if 0	
	gdk_draw_drawable (widget->window,
			widget->style->fg_gc[GTK_WIDGET_STATE (widget)],
			taho->bgpix,
			/* Only copy the area that was exposed. */
			0, 0,
			0, 0,
			widget->allocation.width, widget->allocation.height);
#endif
}

static void 
gtk_taho_size_request(GtkWidget      *widget,
				      GtkRequisition *requisition)
{
	requisition->width = TAHO_DEFAULT_WIDTH;
	requisition->height = TAHO_DEFAULT_HEIGHT;
}

static void
gtk_taho_size_allocate(GtkWidget     *widget,
					   GtkAllocation *allocation)
{
	GtkTaho *taho = GTK_TAHO(widget);

	g_return_if_fail (widget != NULL);
	g_return_if_fail (GTK_IS_TAHO (widget));
	g_return_if_fail (allocation != NULL);

	widget->allocation = *allocation;
	
	if (GTK_WIDGET_REALIZED(widget)) {
		gdk_window_move_resize(widget->window,
				allocation->x, allocation->y,
				allocation->width, allocation->height);
	}

	taho->radius = MIN(allocation->width - LEFT_OFFSET - RIGHT_OFFSET, allocation->height - TOP_OFFSET - BUTTOM_OFFSET);
	taho->xc = (allocation->width - RIGHT_OFFSET);
	taho->yc = (allocation->height - BUTTOM_OFFSET);

	if (widget->window == NULL)
		return;

	if (taho->bgpix != NULL)  {
		gdk_window_set_back_pixmap(widget->window, NULL, FALSE);
		g_object_unref(taho->bgpix);
	}
	
	taho->bgpix = gdk_pixmap_new(widget->window, 
			widget->allocation.width, widget->allocation.height,
			-1);

	draw_taho_bg_pix(taho);
	gdk_window_set_back_pixmap(widget->window, taho->bgpix, FALSE);

	draw_taho_dynamic(taho);
}

static gboolean
gtk_taho_expose(GtkWidget      *widget,
				GdkEventExpose *event )
{
	GtkTaho *taho = GTK_TAHO(widget);
	
	/*
	g_print("expose event\n");
	*/
#if 0 /* Removed because of optimization */
	g_return_val_if_fail (widget != NULL, FALSE);
	g_return_val_if_fail (GTK_IS_TAHO (widget), FALSE);
	g_return_val_if_fail (event != NULL, FALSE);
#endif

	if (event->count > 0)
	    return (FALSE);

#if 0
	gdk_draw_drawable (widget->window,
			widget->style->fg_gc[GTK_WIDGET_STATE (widget)],
			taho->bgpix,
			/* Only copy the area that was exposed. */
			event->area.x, event->area.y,
			event->area.x, event->area.y,
			event->area.width, event->area.height);
#endif

	draw_taho_dynamic(taho);

	return (FALSE);
}

static void
adj_changed(GtkAdjustment *adj, gpointer arg)
{
	GtkTaho *t = (GtkTaho *) arg;

	g_return_if_fail(t != NULL);

	setup_static_text(t);
}

static void
adj_val_changed(GtkAdjustment *adj, gpointer arg)
{
	GtkTaho *t = (GtkTaho *) arg;
	/*
	GdkRectangle rect;
	*/

	/* Optimized, may be not safe.
	g_return_if_fail(t != NULL);
	*/
	
	/*
	g_print("new val: %f\n", t->angle);	
	*/

	/*
	t->angle = M_PI  - (M_PI / 2.) / (adj->upper - adj->lower) * adj->value;
	*/

	t->angle = M_PI * (1.  - (0.5) / (adj->upper - adj->lower) * adj->value);
	
#if 0
	rect.x = GTK_WIDGET(t)->allocation.x;
	rect.y = GTK_WIDGET(t)->allocation.y;
	rect.width = GTK_WIDGET(t)->allocation.width;
	rect.height = GTK_WIDGET(t)->allocation.height;
	
	gdk_window_invalidate_rect(GTK_WIDGET(t)->window, &rect, TRUE);
#else
	gtk_widget_queue_draw(GTK_WIDGET(t));
#endif
	
}

static  void
remove_adj(GtkTaho *t)
{
	g_return_if_fail(t != NULL);
	g_return_if_fail(t->adj != NULL);
	
	g_signal_handlers_disconnect_by_func( G_OBJECT(t->adj),
			G_CALLBACK(adj_changed), 
			t);
	g_signal_handlers_disconnect_by_func( G_OBJECT(t->adj),
			G_CALLBACK(adj_val_changed), 
			t);

	g_object_unref(t->adj);
}


